// -*- C++ -*-
/*!
 * @file  TimedCameraImageViewer.cpp
 * @brief Camara Image Viewer Module
 * @date $Date$
 *
 * $Id$
 */

#include <rtm/SystemLogger.h>
#include "TimedCameraImageViewer.h"

// Module specification
// <rtc-template block="module_spec">
static const char* timedcameraimageviewer_spec[] =
  {
    "implementation_id", "TimedCameraImageViewer",
    "type_name",         "TimedCameraImageViewer",
    "description",       "Camara Image Viewer Module",
    "version",           "0.2.1",
    "vendor",            "Tsuyoshi Suenaga, NAIST",
    "category",          "viewer",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.frames_num", "0",
    // Widget
    "conf.__widget__.frames_num", "text",
    // Constraints
    "conf.__constraints__.frames_num", "x>=0",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
TimedCameraImageViewer::TimedCameraImageViewer(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_imageIn("Image", m_image),
    m_CameraCaptureServicePortPort("CameraCaptureServicePort")

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
TimedCameraImageViewer::~TimedCameraImageViewer()
{
}



RTC::ReturnCode_t TimedCameraImageViewer::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("Image", m_imageIn);
  
  // Set OutPort buffer
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  m_CameraCaptureServicePortPort.registerConsumer("CameraCaptureService", "Img::CameraCaptureService", m_CameraCaptureService);
  
  // Set CORBA Service Ports
  addPort(m_CameraCaptureServicePortPort);
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("frames_num", m_frames_num, "0");
  
  // </rtc-template>
  return RTC::RTC_OK;
}


RTC::ReturnCode_t TimedCameraImageViewer::onFinalize()
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t TimedCameraImageViewer::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t TimedCameraImageViewer::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * サービスポートを通して、画像取得方法の設定を行う。
 */

RTC::ReturnCode_t TimedCameraImageViewer::onActivated(RTC::UniqueId ec_id)
{
  PortServiceList* portlist;

  portlist = this->get_ports();
  port_connect_check = new bool [portlist->length()];
  for(unsigned int i=0; i<portlist->length(); ++i){
    PortService_ptr port;
    port = (*portlist)[i];
    if(port->get_port_profile()->connector_profiles.length()!=0){
      port_connect_check[i]=true;
    }else{
      port_connect_check[i]=false;
    }
  }

  // サービスポート接続状態のチェック
  if(port_connect_check[1]==true){
    // 相手がActivatedか？もチェックすべき
    if(m_frames_num == 0){
      std::cout << "start continuous." << std::endl;
      RTC_INFO(("Send command of \"start continuous\" via CameraCaptureService."));
      m_CameraCaptureService->start_continuous();
    }else if(m_frames_num == 1){
      std::cout << "take one frame." << std::endl;
      RTC_INFO(("Send command of \"take one frame\" via CameraCaptureService."));
      m_CameraCaptureService->take_one_frame();
    }else if(m_frames_num > 1){
      std::cout << "take multi frames (" << m_frames_num << " frames)." << std::endl;
      RTC_INFO(("Send command of \"take multi frames\" via CameraCaptureService."));
      m_CameraCaptureService->take_multi_frames(m_frames_num);
    }else{
      std::cerr << "frames_num >= 0" << std::endl;
      RTC_ERROR(("Configuration Param <frames_num> should be over 0. [%d]",
                 m_frames_num));
      return RTC::RTC_ERROR;
    }
  }

  cv::namedWindow("viewer", CV_WINDOW_AUTOSIZE);
  std::cout << "Viewer start." << std::endl;

  return RTC::RTC_OK;
}

/*!
 * サービスポートを通して、画像取得方法のリセットを行う
 * 。
 */

RTC::ReturnCode_t TimedCameraImageViewer::onDeactivated(RTC::UniqueId ec_id)
{
  if(port_connect_check[1]==true){
    // 相手がActivatedか？もチェックすべき
    if(m_frames_num == 0){
      std::cout << "stop continuous." << std::endl;
      RTC_INFO(("Send command of \"stop continuous\" via CameraCaptureService."));
      m_CameraCaptureService->stop_continuous();
    }
  }
  delete port_connect_check;

#if CV_MAJOR_VERSION == 2 && CV_MINOR_VERSION >= 2
  cv::destroyWindow("viewer");
#endif // CV_MAJOR_VERSION, CV_MINOR_VERSION
  std::cout << "Viewer stop." << std::endl;

  return RTC::RTC_OK;
}

/*!
 * カメラ画像を取得して、画像を描画する．
 */

RTC::ReturnCode_t TimedCameraImageViewer::onExecute(RTC::UniqueId ec_id)
{
  if(m_imageIn.isNew()){
    m_imageIn.read();

    fprintf(stderr, "tm.sec %ld\n", m_image.tm.sec);
    fprintf(stderr, "tm.nsec %ld\n", m_image.tm.nsec);
    
    width = m_image.data.image.width;
    height = m_image.data.image.height;
    channels = (m_image.data.image.format == CF_GRAY) ? 1 :
      (m_image.data.image.format == CF_RGB) ? 3 :
      (m_image.data.image.raw_data.length()/width/height);
    RTC_TRACE(("Capture image size %d x %d", width, height));
    RTC_TRACE(("Channels %d", channels));

    if(channels == 3)
      image.create(height, width, CV_8UC3);
    else
      image.create(height, width, CV_8UC1);

    for(int i=0; i<height; ++i){
      memcpy(&image.data[i*image.step],
             &m_image.data.image.raw_data[i*width*channels],
             sizeof(unsigned char)*width*channels);
    }

    fprintf(stderr, "Intrinsic matrix element\n");
    for(int i=0; i<5; ++i){
      fprintf(stderr, " %f",
              m_image.data.intrinsic.matrix_element[i]);
      RTC_TRACE(("Intrinsic matrix element[%d] %f",
                 i, m_image.data.intrinsic.matrix_element[i]));
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "Distortion parameter\n");
    for(unsigned int i=0;
        i<m_image.data.intrinsic.distortion_coefficient.length(); ++i){
      fprintf(stderr, " %f",
              m_image.data.intrinsic.distortion_coefficient[i]);
      RTC_TRACE(("Distortion parameter[%d] %f",
                 i, m_image.data.intrinsic.distortion_coefficient[i]));
    }
    fprintf(stderr, "\n");

    if(channels == 3)
      cv::cvtColor(image, image, CV_RGB2BGR);
  }

  if(!image.empty())
    cv::imshow("viewer", image);
  cv::waitKey(5);

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t TimedCameraImageViewer::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t TimedCameraImageViewer::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t TimedCameraImageViewer::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t TimedCameraImageViewer::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t TimedCameraImageViewer::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void TimedCameraImageViewerInit(RTC::Manager* manager)
  {
    coil::Properties profile(timedcameraimageviewer_spec);
    manager->registerFactory(profile,
                             RTC::Create<TimedCameraImageViewer>,
                             RTC::Delete<TimedCameraImageViewer>);
  }
  
};


